/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#if defined(__XC)
#include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
#include <htc.h>        /* HiTech General Include File */
#endif

#include <stdint.h>         /* For uint8_t definition */
#include <stdbool.h>
#include <string.h>
#include <pic16f18324.h>

#include "user.h"


/******************************************************************************/
/* User Functions                                                             */

/******************************************************************************/

/**
 * @brief __init_mssp
 * Initialize I2C module
 */
static void __init_mssp(void) {
    SSP1STAT = 0x0; // I2C running at 100kHz
    SSP1CON1 = 0x36; // i2c slave mode, 7-bit addressing without START and STOP bit interrupts

    SSP1CON2bits.SEN = 1; // Enable clock stretching 

    SSP1CON3bits.BOEN = 1; // Enable buffer overflow
    SSP1CON3bits.SDAHT = 1; // Enable 300 ns SDA hold time
    SSP1CON3bits.SBCDE = 1; // Enable bus collision interrupt

    SSPADD = registers.current_settings.address << 1;

}

/**
 * @brief __unlock_pps
 * Unlock sequence for PPS module
 */
static inline void __unlock_pps(void) {
    /* Disable GIE */
    INTCONbits.GIE = 0;

    /* Send magic sequence */
    PPSLOCK = 0x55;
    PPSLOCK = 0xAA;

    /* Unlock PPS */
    PPSLOCKbits.PPSLOCKED = 0;

    /* Enable GIO */
    INTCONbits.GIE = 1;
}

/**
 * @brief __lock_pps
 * Lock sequence for PPS module
 */
static inline void __lock_pps(void) {
    /* Disable GIE */
    INTCONbits.GIE = 0;

    /* Send magic sequence */
    PPSLOCK = 0x55;
    PPSLOCK = 0xAA;

    /* Lock PPS */
    PPSLOCKbits.PPSLOCKED = 1;

    /* Enable GIO */
    INTCONbits.GIE = 1;
}

/**
 * @brief __init_pps
 * Initialize Peripheral Module Disable
 */
static void __init_pps(void) {
    /* Unlock PPS */
    __unlock_pps();

    /* Configure I2C pins */
    SSP1CLKPPSbits.SSP1CLKPPS = 0x02; // SCL on RA2
    RA2PPSbits.RA2PPS = 0x18;

    SSP1DATPPSbits.SSP1DATPPS = 0x04; // SDA on RA4
    RA4PPSbits.RA4PPS = 0x19;

#if !defined(__DEBUG__)

    if (registers.current_settings.mode == MODE_CLC) {
        /* Configure CLC */
        CLCIN0PPSbits.CLCIN0PPS = 0x05; // RA5
        RC5PPSbits.RC5PPS = 0x04; // CLC1OUT

        CLCIN1PPSbits.CLCIN1PPS = 0x14; // RC4
        RC0PPSbits.RC0PPS = 0x05; // CLC2OUT

    } else if (registers.current_settings.mode == MODE_UART) {
        /* Configure EUSART */
        RXPPSbits.RXPPS = 0x14; // RX -> RC4
        RC5PPSbits.RC5PPS = 0x14; // TX -> RC5        

    }



#else
    RC0PPSbits.RC0PPS = 0x14;
#endif
    /* Lock PPS */
    __lock_pps();

}

/**
 * @brief __init_clc
 * Initialize CLC logic
 */
static void __init_clc(void) {

    /* Disable CLC */
    CLC1CONbits.LC1EN = 0;
    CLC2CONbits.LC2EN = 0;
    
    /* Select desired inputs */
    CLC1SEL0bits.D1S = 0; // CLCIN0PPS
    CLC2SEL0bits.D1S = 1; // CLCIN1PPS

    CLC1POLbits.LC1G1POL = 0;
    CLC1POLbits.LC1G2POL = 1;
    CLC1POLbits.LC1G3POL = 1;
    CLC1POLbits.LC1G4POL = 1;

    CLC2POLbits.LC2G1POL = 0;
    CLC2POLbits.LC2G2POL = 1;
    CLC2POLbits.LC2G3POL = 1;
    CLC2POLbits.LC2G4POL = 1;

    CLC1GLS0 = 0x02;
    CLC1GLS1 = 0x00;
    CLC1GLS2 = 0x00;
    CLC1GLS3 = 0x00;

    CLC2GLS0 = 0x02;
    CLC2GLS1 = 0x00;
    CLC2GLS2 = 0x00;
    CLC2GLS3 = 0x00;
    
    CLC1CONbits.LC1MODE = 0x02; // 4-input AND
    CLC2CONbits.LC2MODE = 0x02; // 4-input AND

    /* Enable CLC */
    CLC1CONbits.LC1EN = registers.current_settings.clc_control.tx_en;
    CLC2CONbits.LC2EN = registers.current_settings.clc_control.rx_en;
}

static void __init_gpio(void) {
    ANSELA = 0;
    ANSELC = 0;

    TRISA = 0;
    TRISC = 0;
    
    /* LED1 and LED2 */
    LATAbits.LATA0 = 0;
    LATAbits.LATA1 = 0;

    TRISAbits.TRISA2 = 1; // SCL
    TRISAbits.TRISA4 = 1; // SDA
    
    TRISAbits.TRISA5 = 1; // CLC1IN0
    TRISCbits.TRISC4 = 1; // CLC2IN0

    TRISCbits.TRISC5 = 0; // CLC1
    TRISCbits.TRISC0 = 0; // CLC2

    TRISCbits.TRISC1 = 1; // PROG
    WPUCbits.WPUC1 = 1;

    /* Configure #RE and DE pins */
    LATCbits.LATC2 = !registers.current_settings.clc_control.rx_en;
    LATCbits.LATC3 = registers.current_settings.clc_control.tx_en;
    
    LATAbits.LATA0 = registers.current_settings.clc_control.tx_en;
    LATAbits.LATA1 = registers.current_settings.clc_control.rx_en;
    

}

static void __init_eusart(void) {

    /* Enable WP on RX pin */
    WPUCbits.WPUC4 = 1;
    
    update_baudrate();


    TX1STAbits.TXEN = registers.current_settings.clc_control.tx_en;
    
    if(registers.current_settings.clc_control.rx_en) {
        RC1STAbits.CREN = 1;
        PIE1bits.RCIE = 1;
    } else {
        RC1STAbits.CREN = 0;
        PIE1bits.RCIE = 0;
    }  
}

void InitApp(void) {


    /* Read current configuration */
    read_configuration(&registers.current_settings);
    memcpy(&last_settings, &registers.current_settings, sizeof (last_settings));

    /* Check if EEPROM is empty */
    if (registers.current_settings.address < 0x03 || registers.current_settings.address > 0x77) {
        save_configuration(&default_settings);

        /* Wait for WDT reset */
        while (1) {
            asm("NOP");
        }
    }

    __init_gpio();
#if defined(__DEBUG__)
    __init_eusart();
#endif


    switch (registers.current_settings.mode) {
        case MODE_CLC:
            __init_clc();
            break;
        case MODE_UART:
#if !defined(__DEBUG__)
            __init_eusart();
#endif
            break;
        default:
            /* Invalid mode */
            save_configuration(&default_settings);
            
            /* Waif for WDT reset */
            while (1) {
                asm("NOP");
            }
            break;
    }

    __init_pps();
    __init_mssp();

    PIR1bits.SSP1IF = 0;
    PIE1bits.SSP1IE = 1;

    PIR1bits.BCL1IF = 0;
    PIE1bits.BCL1IE = 1;
    
    PIR1bits.RCIF = 0;

    PEIE = 1;
    GIE = 1;
}

/**
 * @brief 
 * @param c
 */
void uartSend(char c) {
    while (!TX1STAbits.TRMT);
    TX1REG = c;
}

#if defined(__DEBUG__)

void putch(char c) {
    uartSend(c);
}
#endif

/**
 * 
 */
void update_baudrate()
{
    switch(registers.current_settings.baudrate) {
        case UART_B50:          SP1BRG = 39999;     break;
        case UART_B75:          SP1BRG = 26666;     break;
        case UART_B110:         SP1BRG = 18181;     break;
        case UART_B134:         SP1BRG = 14924;     break;
        case UART_B150:         SP1BRG = 13332;     break;
        case UART_B300:         SP1BRG = 6666;      break;
        case UART_B600:         SP1BRG = 3332;      break;
        case UART_B1200:        SP1BRG = 1666;      break;
        case UART_B1800:        SP1BRG = 1110;      break;
        case UART_B2400:        SP1BRG = 832;       break;
        case UART_B4800:        SP1BRG = 416;       break;
        case UART_B7200:        SP1BRG = 277;       break;
        case UART_B9600:        SP1BRG = 207;       break;
        case UART_B14400:       SP1BRG = 138;       break;
        case UART_B19200:       SP1BRG = 103;       break;
        case UART_B38400:       SP1BRG = 51;        break;
        case UART_B57600:       SP1BRG = 34;        break;
        case UART_B76800:       SP1BRG = 25;        break;
        case UART_B115200:      SP1BRG = 68;        break;
        case UART_B128000:      SP1BRG = 62;        break;
        case UART_B230400:      SP1BRG = 34;        break;
        case UART_B500000:      SP1BRG = 15;        break;
        case UART_B576000:      SP1BRG = 13;        break;
        case UART_B1000000:     SP1BRG = 7;         break;
    }

    TX1STAbits.SYNC = 0;
    RC1STAbits.SPEN = 1;
    BRG16 = 1;
    BRGH = (registers.current_settings.baudrate > UART_B76800) ? 1 : 0;
    
}