/******************************************************************************/
/* User Level #define Macros                                                  */
/******************************************************************************/

#define DEFAULT_ADDRESS 0x22
#define DEFAULT_ID      0x25
#define FIRMWARE        0x03

/* Read only registers */
#define REGISTER_ID     0x20        // Device ID: RO
#define REGISTER_FW     0x21        // Device FW version: RO

#define REGISTER_ADDR   0x22
#define REGISTER_MODE   0x23
#define REGISTER_CTRL   0x24
#define REGISTER_BR     0x25

#define REGISTER_TX     0x26
#define REGISTER_RX     0x27

#define _XTAL_FREQ      32000000UL

#define CONFIGS_ADDRESS 0x7000
#define FIFO_SIZE       64

/******************************************************************************/
/* User Function Prototypes                                                   */
/******************************************************************************/

/**
 * @brief mode_t
 * Defines types of operation of the device
 */
typedef enum {
    MODE_CLC = 0, MODE_UART
}mode_t;

/**
 * @brief clc_control_t
 * Used to enable/disable channel
 */
typedef union {
    uint8_t clc;
    struct {
            unsigned rx_en  : 1;
            unsigned tx_en  : 1;
            unsigned        : 6;
        };
}clc_control_t;

typedef enum {
    UART_B50 = 0, 
    UART_B75, 
    UART_B110,
    UART_B134,
    UART_B150,
    UART_B300,
    UART_B600,
    UART_B1200,
    UART_B1800,
    UART_B2400,
    UART_B4800,
    UART_B7200,
    UART_B9600,
    UART_B14400,
    UART_B19200,
    UART_B38400,
    UART_B57600,
    UART_B76800,
    UART_B115200,
    UART_B128000,
    UART_B230400,
    UART_B500000,
    UART_B576000,
    UART_B1000000
} uart_baudrate_t;
/**
 * @brief settings_t
 * Map settings memory
 */
typedef struct {
    uint8_t address;
    mode_t mode;           // UART or I2C
    clc_control_t clc_control;
    uart_baudrate_t baudrate;
}settings_t;

/**
 * @brief registers_t
 * Hold register-like information
 */
typedef struct {
    uint8_t id;
    uint8_t fw;
    settings_t current_settings;
    uint8_t tx;
    uint8_t rx;
}registers_t;


typedef struct {
    int head;
    int tail;
    uint8_t fifo[FIFO_SIZE];
}fifo_t;



void InitApp(void);         /* I/O and Peripheral Initialization */

/* ----------- EUSART functions ---------------*/
void uartSend(char c);
void update_baudrate();

/* ----------- Queue functions --------------*/
extern uint8_t enqueue(fifo_t *fifo, uint8_t data);
extern uint8_t dequeue(fifo_t *fifo);
extern uint8_t queue_available(fifo_t *fifo);

/* ----------- Configuration functions ------------*/
void read_configuration(void *settings);
void save_configuration(const void * const settings);
void erase_configuration(void);

extern registers_t registers;
extern settings_t default_settings;
extern settings_t last_settings;

extern fifo_t rx;
extern fifo_t tx;
